﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.ConditionActionEncoding {
  [Item("XCSEvaluator", "Evaluator tries to match condition and action of IMatching.")]
  [StorableClass]
  public class XCSEvaluator : SingleSuccessorOperator, IXCSEvaluator {
    #region parameter properties
    public IValueLookupParameter<DoubleValue> FitnessParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["Fitness"]; }
    }
    public IValueLookupParameter<DoubleValue> CurrentPayoffParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["CurrentPayoff"]; }
    }
    public IValueLookupParameter<DoubleValue> PredictionParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["Prediction"]; }
    }
    public IValueLookupParameter<DoubleValue> ErrorParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["Error"]; }
    }
    public IValueLookupParameter<IntValue> ExperienceParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["Experience"]; }
    }
    public IValueLookupParameter<IntValue> TimestampParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["Timestamp"]; }
    }
    public IValueLookupParameter<DoubleValue> AverageActionSetSizeParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["AverageActionSetSize"]; }
    }
    public IValueLookupParameter<IntValue> NumerosityParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["Numerosity"]; }
    }
    public IValueLookupParameter<BoolValue> InsertInPopulationParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["InsertInPopulation"]; }
    }
    public ILookupParameter<DoubleValue> InitialPredictionParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["InitialPrediction"]; }
    }
    public ILookupParameter<DoubleValue> InitialErrorParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["InitialError"]; }
    }
    public ILookupParameter<DoubleValue> InitialFitnessParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["InitialFitness"]; }
    }
    public ILookupParameter<PercentValue> BetaParameter {
      get { return (ILookupParameter<PercentValue>)Parameters["Beta"]; }
    }
    public ILookupParameter<PercentValue> AlphaParameter {
      get { return (ILookupParameter<PercentValue>)Parameters["Alpha"]; }
    }
    public ILookupParameter<DoubleValue> PowerParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["v"]; }
    }
    public ILookupParameter<IntValue> ActualTimeParameter {
      get { return (ILookupParameter<IntValue>)Parameters["ActualTime"]; }
    }
    public ILookupParameter<IntValue> CurrentActionSetSizeParameter {
      get { return (ILookupParameter<IntValue>)Parameters["CurrentActionSetSize"]; }
    }
    public ILookupParameter<DoubleValue> ErrorZeroParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["ErrorZero"]; }
    }
    public ILookupParameter<DoubleValue> AccuracyParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Accuracy"]; }
    }
    public ILookupParameter<DoubleValue> AccuracySumParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["AccuracySum"]; }
    }
    #endregion

    #region properties
    private int Experience {
      get { return ExperienceParameter.ActualValue.Value; }
      set { ExperienceParameter.ActualValue.Value = value; }
    }
    private double Beta {
      get { return BetaParameter.ActualValue.Value; }
      set { BetaParameter.ActualValue.Value = value; }
    }
    private double Prediction {
      get { return PredictionParameter.ActualValue.Value; }
      set { PredictionParameter.ActualValue.Value = value; }
    }
    private double CurrentPayoff {
      get { return CurrentPayoffParameter.ActualValue.Value; }
      set { CurrentPayoffParameter.ActualValue.Value = value; }
    }
    private double Error {
      get { return ErrorParameter.ActualValue.Value; }
      set { ErrorParameter.ActualValue.Value = value; }
    }
    private double AverageActionSetSize {
      get { return AverageActionSetSizeParameter.ActualValue.Value; }
      set { AverageActionSetSizeParameter.ActualValue.Value = value; }
    }
    private int CurrentActionSetSize {
      get { return CurrentActionSetSizeParameter.ActualValue.Value; }
      set { CurrentActionSetSizeParameter.ActualValue.Value = value; }
    }
    private double Fitness {
      get { return FitnessParameter.ActualValue.Value; }
      set { FitnessParameter.ActualValue.Value = value; }
    }
    private double Accuracy {
      get { return AccuracyParameter.ActualValue.Value; }
    }
    private double AccuracySum {
      get { return AccuracySumParameter.ActualValue.Value; }
    }
    private int Numerosity {
      get { return NumerosityParameter.ActualValue.Value; }
    }
    #endregion

    [StorableConstructor]
    protected XCSEvaluator(bool deserializing) : base(deserializing) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new XCSEvaluator(this, cloner);
    }
    protected XCSEvaluator(XCSEvaluator original, Cloner cloner)
      : base(original, cloner) {
    }
    public XCSEvaluator()
      : base() {
      Parameters.Add(new ValueLookupParameter<DoubleValue>("Fitness", ""));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("CurrentPayoff", ""));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("Prediction", ""));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("Error", ""));
      Parameters.Add(new ValueLookupParameter<IntValue>("Experience", ""));
      Parameters.Add(new ValueLookupParameter<IntValue>("Timestamp", ""));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("AverageActionSetSize", ""));
      Parameters.Add(new LookupParameter<DoubleValue>("Accuracy", ""));
      Parameters.Add(new LookupParameter<DoubleValue>("AccuracySum", ""));
      Parameters.Add(new ValueLookupParameter<IntValue>("Numerosity", ""));
      Parameters.Add(new ValueLookupParameter<BoolValue>("InsertInPopulation", ""));

      Parameters.Add(new LookupParameter<DoubleValue>("InitialPrediction"));
      Parameters.Add(new LookupParameter<DoubleValue>("InitialError"));
      Parameters.Add(new LookupParameter<DoubleValue>("InitialFitness"));
      Parameters.Add(new LookupParameter<IntValue>("ActualTime"));
      Parameters.Add(new LookupParameter<PercentValue>("Beta"));
      Parameters.Add(new LookupParameter<PercentValue>("Alpha"));
      Parameters.Add(new LookupParameter<DoubleValue>("v"));
      Parameters.Add(new LookupParameter<DoubleValue>("ErrorZero"));
      Parameters.Add(new LookupParameter<IntValue>("CurrentActionSetSize"));
    }

    public override IOperation Apply() {
      if (PredictionParameter.ActualValue == null) {
        //the classifier has just been created
        PredictionParameter.ActualValue = new DoubleValue(InitialPredictionParameter.ActualValue.Value);
        ErrorParameter.ActualValue = new DoubleValue(InitialErrorParameter.ActualValue.Value);
        FitnessParameter.ActualValue = new DoubleValue(InitialFitnessParameter.ActualValue.Value);
        ExperienceParameter.ActualValue = new IntValue(0);
        TimestampParameter.ActualValue = new IntValue(ActualTimeParameter.ActualValue.Value);
        AverageActionSetSizeParameter.ActualValue = new DoubleValue(1);
        NumerosityParameter.ActualValue = new IntValue(1);
        InsertInPopulationParameter.ActualValue = new BoolValue(false);
      } else {
        Experience += 1;
        if (Experience < (1 / Beta)) {
          Prediction += (CurrentPayoff - Prediction) / Experience;
          Error += (Math.Abs(CurrentPayoff - Prediction) - Error) / Experience;
          AverageActionSetSize += (CurrentActionSetSize - AverageActionSetSize) / Experience;
        } else {
          Prediction += Beta * (CurrentPayoff - Prediction);
          Error += Beta * (Math.Abs(CurrentPayoff - Prediction) - Error);
          AverageActionSetSize += Beta * (CurrentActionSetSize - AverageActionSetSize);
        }
        Fitness += Beta * ((Accuracy * Numerosity) / AccuracySum - Fitness);
      }
      return base.Apply();
    }
  }
}
